package sysbench;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

import org.apache.commons.dbcp.BasicDataSource;

public class Manager {
	private final Config config;
	private final AtomicInteger transactionCount = new AtomicInteger();
	private final CountDownLatch progressLatch = new CountDownLatch(1);
	private final BasicDataSource dataSource = new BasicDataSource();

	private volatile int referenceCount;
	private volatile boolean status = true;

	public static void main(String[] args) {
		try {
			System.out.println("Begin : "
					+ new SimpleDateFormat("yyyy/MM/dd-HH:mm:ss").format(new Date()));
			Config config = new Config(args);
			System.out.println(config.getConfigString());
			Manager manager = new Manager(config);
			manager.measure();
			System.out.println("End : "
					+ new SimpleDateFormat("yyyy/MM/dd-HH:mm:ss").format(new Date()));
		} catch (Exception e) {
			e.printStackTrace(System.out);
		}
	}

	public Manager(Config config) {
		this.config = config;
	}

	public void measure() {
		if (config.useConnectionPool()) {
			createDataSource();
		}

		List<Worker> workerList = new ArrayList<Worker>();

		for (int workerId = 0; workerId < config.getNumWorkers(); workerId++) {
			Worker worker = null;

			if (config.usePreparedStatement()) {
				worker = new WorkerImplBind(this, workerId, config.getOltpTableSize(), config
						.getOltpRangeSize());
			} else {
				worker = new WorkerImplLiteral(this, workerId, config.getOltpTableSize(), config
						.getOltpRangeSize());
			}

			workerList.add(worker);
		}

		ScheduledExecutorService progress = Executors.newSingleThreadScheduledExecutor();
		progress.scheduleAtFixedRate(new Progress(), 0L, 1L, TimeUnit.SECONDS);

		for (Worker worker : workerList) {
			worker.start();
		}

		try {
			progressLatch.await();
		} catch (InterruptedException e) {
		}

		this.status = false;
		progress.shutdownNow();

		for (Worker worker : workerList) {
			do {
				try {
					worker.getThread().join();
				} catch (InterruptedException e) {
				}
			} while (worker.getThread().isAlive());
		}

		System.out
				.println("Throughput : "
						+ ((double) (transactionCount.get() - referenceCount) / config
								.getMeasurementTime()));
	}

	public Connection getConnection() throws SQLException {
		Connection connection = null;

		if (config.useConnectionPool()) {
			connection = dataSource.getConnection();
		} else {
			connection = DriverManager.getConnection(config.getJdbcUrl(), config.getJdbcUser(),
					config.getJdbcPass());
			connection.setAutoCommit(false);
		}

		return connection;
	}

	public boolean getStatus() {
		return status;
	}

	public void incrementTransactionCount() {
		transactionCount.incrementAndGet();
	}

	public synchronized void putException(int workerId, Exception e) {
		System.out.println(workerId);
		e.printStackTrace(System.out);
		progressLatch.countDown();
	}

	private void createDataSource() {
		dataSource.setUrl(config.getJdbcUrl());
		dataSource.setUsername(config.getJdbcUser());
		dataSource.setPassword(config.getJdbcPass());
		dataSource.setDefaultAutoCommit(false);
		dataSource.setInitialSize(config.getNumWorkers());
		dataSource.setMaxActive(config.getNumWorkers());
		dataSource.setMaxIdle(-1);

		if (config.usePreparedStatementPool()) {
			dataSource.setPoolPreparedStatements(true);
			dataSource.setMaxOpenPreparedStatements(100);
		} else {
			dataSource.setPoolPreparedStatements(false);
		}
	}

	private class Progress implements Runnable {
		private boolean doPrint = true;
		private int transactionCount;
		private int previousCount;
		private int elapsedTime = 0 - config.getWarmupTime();

		@Override
		public void run() {
			if (doPrint) {
				this.elapsedTime++;
				this.transactionCount = Manager.this.transactionCount.get();
				System.out.println(elapsedTime + " : " + (transactionCount - previousCount));
				this.previousCount = transactionCount;
			}

			if (elapsedTime == 0) {
				Manager.this.referenceCount = transactionCount;
			}

			if (elapsedTime == config.getMeasurementTime()) {
				this.doPrint = false;
				progressLatch.countDown();
			}
		}
	}
}
